@@ -1,10 +1,11 @@
 /* GDA MySQL provider
- * Copyright (C) 1998-2004 The GNOME Foundation.
+ * Copyright (C) 1998 - 2005 The GNOME Foundation.
  *
  * AUTHORS:
  *      Michael Lausch <michael@lausch.at>
  *	Rodrigo Moya <rodrigo@gnome-db.org>
  *      Vivien Malerba <malerba@gnome-db.org>
+ *	Bas Driessen <bas.driessen@xobas.com>
  *
  * This Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -23,6 +24,7 @@
  */
 
 #include <libgda/gda-data-model-array.h>
+#include <libgda/gda-column-index.h>
 #include <libgda/gda-util.h>
 #include <libgda/gda-intl.h>
 #include <stdlib.h>
@@ -32,62 +34,95 @@
 
 #define PARENT_TYPE GDA_TYPE_SERVER_PROVIDER
 
-#define OBJECT_DATA_MYSQL_HANDLE "GDA_Mysql_MysqlHandle"
-
 static void gda_mysql_provider_class_init (GdaMysqlProviderClass *klass);
 static void gda_mysql_provider_init       (GdaMysqlProvider *provider,
 					   GdaMysqlProviderClass *klass);
 static void gda_mysql_provider_finalize   (GObject *object);
 
 static const gchar *gda_mysql_provider_get_version (GdaServerProvider *provider);
+
 static gboolean gda_mysql_provider_open_connection (GdaServerProvider *provider,
 						    GdaConnection *cnc,
 						    GdaQuarkList *params,
 						    const gchar *username,
 						    const gchar *password);
+
 static gboolean gda_mysql_provider_close_connection (GdaServerProvider *provider,
 						     GdaConnection *cnc);
+
 static const gchar *gda_mysql_provider_get_server_version (GdaServerProvider *provider,
 							   GdaConnection *cnc);
+
 static const gchar *gda_mysql_provider_get_database (GdaServerProvider *provider,
 						     GdaConnection *cnc);
+
 static gboolean gda_mysql_provider_change_database (GdaServerProvider *provider,
 		                                    GdaConnection *cnc,
 		                                    const gchar *name);
+
 static gboolean gda_mysql_provider_create_database (GdaServerProvider *provider,
 						    GdaConnection *cnc,
 						    const gchar *name);
+
 static gboolean gda_mysql_provider_drop_database (GdaServerProvider *provider,
 						  GdaConnection *cnc,
 						  const gchar *name);
+
+static gboolean gda_mysql_provider_create_table (GdaServerProvider *provider,
+                                                    GdaConnection *cnc,
+                                                    const gchar *table_name,
+                                                    const GList *attributes_list,
+                                                    const GList *index_list);
+
+static gboolean gda_mysql_provider_drop_table (GdaServerProvider *provider,
+                                                     GdaConnection *cnc,
+                                                     const gchar *table_name);
+
+static gboolean gda_mysql_provider_create_index (GdaServerProvider *provider,
+                                                    GdaConnection *cnc,
+                                                    const GdaDataModelIndex *index,
+                                                    const gchar *table_name);
+
+static gboolean gda_mysql_provider_drop_index (GdaServerProvider *provider,
+                                                     GdaConnection *cnc,
+                                                     const gchar *index_name,
+						     gboolean primary_key,
+                                                     const gchar *table_name);
+
 static GList *gda_mysql_provider_execute_command (GdaServerProvider *provider,
 						  GdaConnection *cnc,
 						  GdaCommand *cmd,
 						  GdaParameterList *params);
+
 static gchar *gda_mysql_provider_get_last_insert_id (GdaServerProvider *provider,
 						     GdaConnection *cnc,
 						     GdaDataModel *recset);
+
 static gboolean gda_mysql_provider_begin_transaction (GdaServerProvider *provider,
 						      GdaConnection *cnc,
 						      GdaTransaction *xaction);
+
 static gboolean gda_mysql_provider_commit_transaction (GdaServerProvider *provider,
 						       GdaConnection *cnc,
 						       GdaTransaction *xaction);
+
 static gboolean gda_mysql_provider_rollback_transaction (GdaServerProvider *provider,
 							 GdaConnection *cnc,
 							 GdaTransaction *xaction);
+
 static gboolean gda_mysql_provider_supports (GdaServerProvider *provider,
 					     GdaConnection *cnc,
 					     GdaConnectionFeature feature);
+
 static GdaDataModel *gda_mysql_provider_get_schema (GdaServerProvider *provider,
 						    GdaConnection *cnc,
 						    GdaConnectionSchema schema,
 						    GdaParameterList *params);
-static gboolean gda_mysql_provider_escape_string (GdaServerProvider *provider,
+gchar *gda_mysql_provider_value_to_sql_string (GdaServerProvider *provider,
 						  GdaConnection *cnc,
-						  const gchar *from,
-						  gchar *to);
-				 		 
+						  GdaValue *from);
+
+
 static GObjectClass *parent_class = NULL;
 
 /*
@@ -111,6 +146,10 @@ gda_mysql_provider_class_init (GdaMysqlProviderClass *klass)
 	provider_class->change_database = gda_mysql_provider_change_database;
 	provider_class->create_database = gda_mysql_provider_create_database;
 	provider_class->drop_database = gda_mysql_provider_drop_database;
+	provider_class->create_table = gda_mysql_provider_create_table;
+	provider_class->drop_table = gda_mysql_provider_drop_table;
+	provider_class->create_index = gda_mysql_provider_create_index;
+	provider_class->drop_index = gda_mysql_provider_drop_index;
 	provider_class->execute_command = gda_mysql_provider_execute_command;
 	provider_class->get_last_insert_id = gda_mysql_provider_get_last_insert_id;
 	provider_class->begin_transaction = gda_mysql_provider_begin_transaction;
@@ -118,7 +157,7 @@ gda_mysql_provider_class_init (GdaMysqlProviderClass *klass)
 	provider_class->rollback_transaction = gda_mysql_provider_rollback_transaction;
 	provider_class->supports = gda_mysql_provider_supports;
 	provider_class->get_schema = gda_mysql_provider_get_schema;
-	provider_class->escape_string = gda_mysql_provider_escape_string;
+	provider_class->value_to_sql_string = gda_mysql_provider_value_to_sql_string;
 }
 
 static void
@@ -488,6 +527,342 @@ gda_mysql_provider_drop_database (GdaServerProvider *provider,
 	return TRUE;
 }
 
+/* create_table handler for the GdaMysqlProvider class */
+static gboolean
+gda_mysql_provider_create_table (GdaServerProvider *provider,
+					GdaConnection *cnc,
+					const gchar *table_name,
+					const GList *attributes_list,
+					const GList *index_list)
+{
+	MYSQL *mysql;
+	GdaMysqlProvider *myprv = (GdaMysqlProvider *) provider;
+	GdaColumn *dmca;
+	GdaDataModelIndex *dmi;
+	GString *sql;
+	gint i, rc;
+	gchar *mysql_data_type, *default_value, *references;
+	glong size;
+	GdaValueType value_type;
+	gboolean retval;
+
+	g_return_val_if_fail (GDA_IS_MYSQL_PROVIDER (myprv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+	g_return_val_if_fail (attributes_list != NULL, FALSE);
+
+	/* check for valid MySQL handle */
+	mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
+	if (!mysql) {
+		gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
+		return FALSE;
+	}
+
+	/* prepare the SQL command */
+	sql = g_string_new ("CREATE TABLE ");
+	g_string_append_printf (sql, "%s (", table_name);
+
+	/* step through list */
+	for (i=0; i<g_list_length ((GList *) attributes_list); i++) {
+
+		if (i>0)
+			g_string_append_printf (sql, ", ");
+
+		dmca = (GdaColumn *) g_list_nth_data ((GList *) attributes_list, i);
+
+		/* name */
+		g_string_append_printf (sql, "`%s` ", gda_column_get_name (dmca));
+
+		/* data type */
+		value_type = gda_column_get_gdatype (dmca);
+		mysql_data_type = gda_mysql_type_from_gda (value_type);
+		g_string_append_printf (sql, "%s", mysql_data_type);
+		g_free(mysql_data_type);
+
+		/* size */
+		size = gda_column_get_defined_size (dmca);
+		if (value_type == GDA_VALUE_TYPE_STRING)
+			g_string_append_printf (sql, "(%ld)", size);
+		else if (value_type == GDA_VALUE_TYPE_NUMERIC)
+			g_string_append_printf (sql, "(%ld,%ld)",
+				gda_column_get_defined_size (dmca),
+				gda_column_get_scale (dmca));
+		else if (value_type == GDA_VALUE_TYPE_MONEY)
+			g_string_append_printf (sql, "(%ld)", size);
+
+		/* optional sizes */
+		if ((value_type == GDA_VALUE_TYPE_BIGINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_BIGUINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_BIGUINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_DOUBLE) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_INTEGER) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_SINGLE) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_SMALLINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_SMALLUINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_TINYINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_TINYUINT) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+		else if ((value_type == GDA_VALUE_TYPE_UINTEGER) && (size > 0))
+			g_string_append_printf (sql, "(%ld)", size);
+
+		/* set UNSIGNED if applicable */
+		switch (value_type) {
+		case GDA_VALUE_TYPE_BIGUINT :
+			g_string_append_printf (sql, "UNSIGNED");
+		case GDA_VALUE_TYPE_SMALLUINT :
+			g_string_append_printf (sql, "UNSIGNED");
+		case GDA_VALUE_TYPE_TINYUINT :
+			g_string_append_printf (sql, "UNSIGNED");
+		case GDA_VALUE_TYPE_UINTEGER :
+			g_string_append_printf (sql, "UNSIGNED");
+		}
+
+		/* NULL */
+		if (gda_column_get_allow_null (dmca) == TRUE)
+			g_string_append_printf (sql, " NULL");
+		else
+			g_string_append_printf (sql, " NOT NULL");
+
+		/* auto increment */
+		if (gda_column_get_auto_increment (dmca) == TRUE)
+			g_string_append_printf (sql, "AUTO_INCREMENT");
+
+		/* (primary) key */
+		if (gda_column_get_primary_key (dmca) == TRUE)
+			g_string_append_printf (sql, " PRIMARY KEY");
+		else
+			if (gda_column_get_unique_key (dmca) == TRUE)
+				g_string_append_printf (sql, " UNIQUE");
+			
+		/* default value (in case of string, user needs to add "'" around the field) */
+                if (gda_column_get_default_value (dmca) != NULL) {
+			default_value = gda_value_stringify ((GdaValue *) gda_column_get_default_value (dmca));
+			if ((default_value != NULL) && (*default_value != '\0'))
+				g_string_append_printf (sql, " DEFAULT %s", default_value);
+		}
+
+		/* any additional parameters */
+		if (gda_column_get_references (dmca) != NULL) {
+			references = (gchar *) gda_column_get_references (dmca);
+			if ((references != NULL) && (*references != '\0'))
+				g_string_append_printf (sql, " %s", references);
+		}
+	}
+
+	/* finish the SQL command */
+	g_string_append_printf (sql, ")");
+
+	/* execute sql command */
+	rc = mysql_query (mysql, sql->str);
+	if (rc != 0) {
+		gda_connection_add_error (cnc, gda_mysql_make_error (mysql));
+		return FALSE;
+	}
+
+	/* clean up */
+	g_string_free (sql, TRUE);
+
+	/* Create indexes */
+	if (index_list != NULL) {
+
+		/* step through index_list */
+		for (i=0; i<g_list_length ((GList *) index_list); i++) {
+
+			/* get index */
+			dmi = (GdaDataModelIndex *) g_list_nth_data ((GList *) index_list, i);
+
+			/* create index */
+			retval = gda_mysql_provider_create_index (provider, cnc, dmi, table_name);
+			if (retval == FALSE)
+				return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
+/* drop_table handler for the GdaMysqlProvider class */
+static gboolean
+gda_mysql_provider_drop_table (GdaServerProvider *provider,
+				  GdaConnection *cnc,
+				  const gchar *table_name)
+{
+	gint rc;
+	gchar *sql;
+	MYSQL *mysql;
+	GdaMysqlProvider *myprv = (GdaMysqlProvider *) provider;
+
+	g_return_val_if_fail (GDA_IS_MYSQL_PROVIDER (myprv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
+	if (!mysql) {
+		gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
+		return FALSE;
+	}
+
+	sql = g_strdup_printf ("DROP TABLE %s", table_name);
+	rc = mysql_query (mysql, sql);
+	g_free (sql);
+	if (rc != 0) {
+		gda_connection_add_error (cnc, gda_mysql_make_error (mysql));
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+/* create_index handler for the GdaMysqlProvider class */
+static gboolean
+gda_mysql_provider_create_index (GdaServerProvider *provider,
+					GdaConnection *cnc,
+					const GdaDataModelIndex *index,
+					const gchar *table_name)
+{
+	MYSQL *mysql;
+	GdaMysqlProvider *myprv = (GdaMysqlProvider *) provider;
+	GdaColumnIndex *dmcia;
+	GString *sql;
+	GList *col_list;
+	gchar *index_name, *col_ref, *idx_ref;
+	gint i, rc, size;
+	gboolean retval = FALSE;
+	GdaSorting sort;
+
+	g_return_val_if_fail (GDA_IS_MYSQL_PROVIDER (myprv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (index != NULL, FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	/* get MySQL handle */
+	mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
+	if (!mysql) {
+		gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
+		return FALSE;
+	}
+
+	/* prepare the SQL command */
+	sql = g_string_new ("ALTER TABLE ");
+
+	/* get index name */
+	index_name = (gchar *) gda_data_model_index_get_name ((GdaDataModelIndex *) index);
+
+	/* determine type of index (PRIMARY, UNIQUE or INDEX) */
+	if (gda_data_model_index_get_primary_key ((GdaDataModelIndex *) index) == TRUE)
+		g_string_append_printf (sql, "%s ADD PRIMARY KEY (", table_name, index_name);
+	else if (gda_data_model_index_get_unique_key ((GdaDataModelIndex *) index) == TRUE)
+		g_string_append_printf (sql, "%s ADD UNIQUE `%s` (", table_name, index_name);
+	else
+		g_string_append_printf (sql, "%s ADD INDEX `%s` (", table_name, index_name);
+
+	/* get the list of index columns */
+	col_list = gda_data_model_index_get_column_index_list ((GdaDataModelIndex *) index);
+
+	/* step through list */
+	for (i = 0; i < g_list_length ((GList *) col_list); i++) {
+
+		if (i > 0)
+			g_string_append_printf (sql, ", ");
+
+		dmcia = (GdaColumnIndex *) g_list_nth_data ((GList *) col_list, i);
+
+		/* name */
+		g_string_append_printf (sql, "`%s` ", gda_column_index_get_column_name (dmcia));
+
+		/* size */
+		if (gda_column_index_get_defined_size (dmcia) > 0) {
+			size = gda_column_index_get_defined_size (dmcia);
+			g_string_append_printf (sql, " (%d) ", size);
+		}
+
+		/* sorting */
+		sort = gda_column_index_get_sorting (dmcia);
+		if (sort == GDA_SORTING_DESCENDING)
+			g_string_append_printf (sql, " DESC ");
+		else
+			g_string_append_printf (sql, " ASC ");
+
+		/* any additional column parameters */
+		if (gda_column_index_get_references (dmcia) != NULL) {
+			col_ref = (gchar *) gda_column_index_get_references (dmcia);
+			if ((col_ref != NULL) && (*col_ref != '\0'))
+				g_string_append_printf (sql, " %s ", col_ref);
+		}
+	}
+
+	/* finish the SQL column(s) section command */
+	g_string_append_printf (sql, ")");
+
+	/* any additional index parameters */
+	if (gda_data_model_index_get_references ((GdaDataModelIndex *) index) != NULL) {
+		idx_ref = (gchar *) gda_data_model_index_get_references ((GdaDataModelIndex *) index);
+		if ((idx_ref != NULL) && (*idx_ref != '\0'))
+		g_string_append_printf (sql, " %s ", idx_ref);
+	}
+
+        /* execute sql command */
+	rc = mysql_query (mysql, sql->str);
+	if (rc != 0) {
+		gda_connection_add_error (cnc, gda_mysql_make_error (mysql));
+		return FALSE;
+	}
+
+	/* clean up */
+	g_string_free (sql, TRUE);
+
+	return TRUE;
+}
+
+/* drop_index handler for the GdaMysqlProvider class */
+static gboolean
+gda_mysql_provider_drop_index (GdaServerProvider *provider,
+				  GdaConnection *cnc,
+				  const gchar *index_name,
+				  gboolean primary_key,
+				  const gchar *table_name)
+{
+	gint rc;
+	gchar *sql;
+	MYSQL *mysql;
+	GdaMysqlProvider *myprv = (GdaMysqlProvider *) provider;
+
+	g_return_val_if_fail (GDA_IS_MYSQL_PROVIDER (myprv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (index_name != NULL, FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	/* get MySQL handle */
+	mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
+	if (!mysql) {
+		gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
+		return FALSE;
+	}
+
+	if (primary_key == TRUE)
+		sql = g_strdup_printf ("ALTER TABLE %s DROP PRIMARY KEY", table_name);
+	else
+		sql = g_strdup_printf ("ALTER TABLE %s DROP INDEX `%s`", table_name, index_name);
+
+	rc = mysql_query (mysql, sql);
+	g_free (sql);
+	if (rc != 0) {
+		gda_connection_add_error (cnc, gda_mysql_make_error (mysql));
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 /* execute_command handler for the GdaMysqlProvider class */
 static GList *
 gda_mysql_provider_execute_command (GdaServerProvider *provider,
@@ -713,7 +1088,7 @@ add_aggregate_row (GdaDataModelArray *recset, const gchar *str, const gchar *com
 	/* 7th the SQL definition */
 	list = g_list_append (list, gda_value_new_string (NULL));
 
-	gda_data_model_append_row (GDA_DATA_MODEL (recset), list);
+	gda_data_model_append_values (GDA_DATA_MODEL (recset), list);
 
 	g_list_foreach (list, (GFunc) gda_value_free, NULL);
 	g_list_free (list);
@@ -927,14 +1302,14 @@ get_mysql_tables (GdaConnection *cnc, GdaParameterList *params)
 		gchar *str;
 
 		/* 1st, the name of the table */
-		name = gda_value_stringify (gda_data_model_get_value_at (GDA_DATA_MODEL (recset), 0, r));
+		name = gda_value_stringify ((GdaValue *) gda_data_model_get_value_at (GDA_DATA_MODEL (recset), 0, r));
 		value_list = g_list_append (value_list, gda_value_new_string (name));
 
 		/* 2nd, the owner */
 		value_list = g_list_append (value_list, gda_value_new_string (""));
 
 		/* 3rd, the comments */
-		str = gda_value_stringify (gda_data_model_get_value_at (GDA_DATA_MODEL (recset), 14, r));
+		str = gda_value_stringify ((GdaValue *) gda_data_model_get_value_at (GDA_DATA_MODEL (recset), 14, r));
 		value_list = g_list_append (value_list, gda_value_new_string (str));
 		g_free (str);
 
@@ -944,7 +1319,7 @@ get_mysql_tables (GdaConnection *cnc, GdaParameterList *params)
 		g_free (str);
 		if (reclist && GDA_IS_DATA_MODEL (reclist->data)) {
 			str = gda_value_stringify (
-				gda_data_model_get_value_at (GDA_DATA_MODEL (reclist->data), 1, 0));
+				(GdaValue *) gda_data_model_get_value_at (GDA_DATA_MODEL (reclist->data), 1, 0));
 			value_list = g_list_append (value_list, gda_value_new_string (str));
 
 			g_free (str);
@@ -954,7 +1329,7 @@ get_mysql_tables (GdaConnection *cnc, GdaParameterList *params)
 		else
 			value_list = g_list_append (value_list, gda_value_new_string (""));
 
-		gda_data_model_append_row (model, value_list);
+		gda_data_model_append_values (model, value_list);
 
 		g_free (name);
 		g_list_foreach (value_list, (GFunc) gda_value_free, NULL);
@@ -964,6 +1339,89 @@ get_mysql_tables (GdaConnection *cnc, GdaParameterList *params)
 	return model;
 }
 
+static GdaDataModel *
+get_mysql_views (GdaConnection *cnc, GdaParameterList *params)
+{
+	GList *reclist;
+	GdaMysqlRecordset *recset;
+	GdaDataModel *model;
+	gint rows, r;
+
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
+
+	/* add the extra information */
+	model = gda_data_model_array_new (4);
+	gda_data_model_set_column_title (model, 0, _("Name"));
+	gda_data_model_set_column_title (model, 1, _("Owner"));
+	gda_data_model_set_column_title (model, 2, _("Comments"));
+	gda_data_model_set_column_title (model, 3, "SQL");
+
+	/* views are not yet supported in MySql => return an empty set */
+
+	return model;
+}
+
+static GdaDataModel *
+get_mysql_procedures (GdaConnection *cnc, GdaParameterList *params)
+{
+	GList *reclist;
+	GdaDataModel *model;
+	gint rows, r;
+	gint i;
+
+	struct {
+		const gchar *name;
+		const gchar *id;
+		/*const gchar *owner; "" */
+		const gchar *comments;
+		const gchar *rettype;
+		gint         nb_args;
+		const gchar *argtypes;
+		/*const gchar *definition; NULL */
+	} procs[] = {
+		{ "ascii", "ascii", "", "smallint", 1, "text"},
+		{ "bin", "bin", "", "text", 1, "bigint"},
+		{ "bit_length", "bit_length", "", "int", 1, "text"},
+		{ "char", "char", "", "text", -1, "-"}
+	};
+
+
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
+
+	/* add the extra information */
+	model = gda_data_model_array_new (8);
+	gda_data_model_set_column_title (model, 0, _("Procedure"));
+	gda_data_model_set_column_title (model, 1, _("Id"));
+	gda_data_model_set_column_title (model, 2, _("Owner"));
+	gda_data_model_set_column_title (model, 3, _("Comments"));
+	gda_data_model_set_column_title (model, 4, _("Return type"));
+	gda_data_model_set_column_title (model, 5, _("Nb args"));
+	gda_data_model_set_column_title (model, 6, _("Args types"));
+	gda_data_model_set_column_title (model, 7, _("Definition"));
+
+	/* fill the recordset */
+	for (i = 0; i < sizeof (procs) / sizeof (procs[0]); i++) {
+		GList *value_list = NULL;
+
+		value_list = g_list_append (value_list, gda_value_new_string (procs[i].name));
+		value_list = g_list_append (value_list, gda_value_new_string (procs[i].id));
+		value_list = g_list_append (value_list, gda_value_new_string (""));
+		value_list = g_list_append (value_list, gda_value_new_string (procs[i].comments));
+		value_list = g_list_append (value_list, gda_value_new_string (procs[i].rettype));
+		value_list = g_list_append (value_list, gda_value_new_integer (procs[i].nb_args));
+		value_list = g_list_append (value_list, gda_value_new_string (procs[i].argtypes));
+		value_list = g_list_append (value_list, gda_value_new_string (NULL));
+
+		gda_data_model_append_values (GDA_DATA_MODEL (model), value_list);
+
+		g_list_foreach (value_list, (GFunc) gda_value_free, NULL);
+		g_list_free (value_list);
+	}
+
+	return model;
+}
+
+
 static GdaDataModel *
 get_mysql_types (GdaConnection *cnc, GdaParameterList *params)
 {
@@ -974,45 +1432,54 @@ get_mysql_types (GdaConnection *cnc, GdaParameterList *params)
 		const gchar *owner;
 		const gchar *comments;
 		GdaValueType type;
+		const gchar *synonyms;
 	} types[] = {
-		{ "blob", "", "Binary blob", GDA_VALUE_TYPE_BINARY },
-		{ "bigint", "", "Big integer", GDA_VALUE_TYPE_BIGINT },
-		{ "char", "", "Char", GDA_VALUE_TYPE_STRING },
-		{ "date", "", "Date", GDA_VALUE_TYPE_DATE },
-		{ "datetime", "", "Date and time", GDA_VALUE_TYPE_TIMESTAMP },
-		{ "decimal", "", "Decimal number", GDA_VALUE_TYPE_DOUBLE },
-		{ "double", "", "Double precision number", GDA_VALUE_TYPE_DOUBLE },
-		{ "enum", "", "Enumeration", GDA_VALUE_TYPE_STRING },
-		{ "float", "", "Single precision number", GDA_VALUE_TYPE_SINGLE },
-		{ "int", "", "Integer", GDA_VALUE_TYPE_INTEGER },
-		{ "integer", "", "Integer", GDA_VALUE_TYPE_INTEGER },		
-		{ "long", "", "Long integer", GDA_VALUE_TYPE_INTEGER },
-		{ "longblob", "", "Long blob", GDA_VALUE_TYPE_BINARY },
-		{ "longtext", "", "Long text", GDA_VALUE_TYPE_STRING },
-		{ "mediumint", "", "Medium integer", GDA_VALUE_TYPE_INTEGER },
-		{ "mediumblob", "", "Medium blob", GDA_VALUE_TYPE_BINARY },
-		{ "mediumtext", "", "Medium text", GDA_VALUE_TYPE_STRING },				
-		{ "set", "", "Set", GDA_VALUE_TYPE_STRING },
-		{ "smallint", "", "Small integer", GDA_VALUE_TYPE_INTEGER },
-		{ "text", "", "Text", GDA_VALUE_TYPE_STRING },
-		{ "tinyint", "", "Tiny integer", GDA_VALUE_TYPE_INTEGER }, /* really a boolean */
-		{ "tinyblob", "", "Tiny blob", GDA_VALUE_TYPE_BINARY },
-		{ "tinytext", "", "Tiny text", GDA_VALUE_TYPE_STRING },		
-		{ "time", "", "Time", GDA_VALUE_TYPE_TIME },
-		{ "timestamp", "", "Time stamp", GDA_VALUE_TYPE_TIMESTAMP },
-		{ "varchar", "", "Variable Length Char", GDA_VALUE_TYPE_STRING },
-		{ "year", "", "Year", GDA_VALUE_TYPE_INTEGER }
+		{ "bool", "", "Boolean type", GDA_VALUE_TYPE_BINARY, "boolean" },
+		{ "blob", "", "Binary blob (up to 65535 bytes)", GDA_VALUE_TYPE_BINARY, NULL },
+		{ "bigint", "", "Big integer, range is -9223372036854775808 to 9223372036854775807", GDA_VALUE_TYPE_BIGINT, NULL  },
+		{ "bigint unsigned", "", "Big unsigned integer, range is 0 to 18446744073709551615", GDA_VALUE_TYPE_BIGINT, NULL  },
+		{ "char", "", "Char", GDA_VALUE_TYPE_STRING, "binary"  },
+		{ "date", "", "Date", GDA_VALUE_TYPE_DATE, NULL  },
+		{ "datetime", "", "Date and time", GDA_VALUE_TYPE_TIMESTAMP, NULL  },
+		{ "decimal", "", "Decimal number", GDA_VALUE_TYPE_DOUBLE, "dec,numeric,fixed"  },
+		{ "double", "", "Double precision number", GDA_VALUE_TYPE_DOUBLE, "double precision,real"  },
+		{ "double unsigned", "", "Unsigned double precision number", GDA_VALUE_TYPE_DOUBLE, "double precision unsigned,real unsigned"  },
+		{ "enum", "", "Enumeration: a string object that can have only one value, chosen from the list of values 'value1', 'value2', ..., NULL or the special '' error value. An ENUM column can have a maximum of 65,535 distinct values", GDA_VALUE_TYPE_STRING, NULL  },
+		{ "float", "", "Floating point number", GDA_VALUE_TYPE_SINGLE , NULL },
+		{ "float unsigned", "", "Unsigned floating point number", GDA_VALUE_TYPE_SINGLE , NULL },
+		{ "int", "", "Integer, range is -2147483648 to 2147483647", GDA_VALUE_TYPE_INTEGER, "integer"  },
+		{ "int unsigned", "", "Unsigned integer, range is 0 to 4294967295", GDA_VALUE_TYPE_UINTEGER, "integer unsigned"  },
+		{ "long", "", "Long integer", GDA_VALUE_TYPE_INTEGER, NULL  },
+		{ "longblob", "", "Long blob (up to 4Gb)", GDA_VALUE_TYPE_BINARY, NULL  },
+		{ "longtext", "", "Long text (up to 4Gb characters)", GDA_VALUE_TYPE_STRING, NULL  },
+		{ "mediumint", "", "Medium integer, range is -8388608 to 8388607", GDA_VALUE_TYPE_INTEGER, NULL  },
+		{ "mediumint unsigned", "", "Medium unsigned integer, range is 0 to 16777215", GDA_VALUE_TYPE_INTEGER, NULL  },
+		{ "mediumblob", "", "Medium blob (up to 16777215 bytes)", GDA_VALUE_TYPE_BINARY, NULL  },
+		{ "mediumtext", "", "Medium text (up to 16777215 characters)", GDA_VALUE_TYPE_STRING, NULL  },				
+		{ "set", "", "Set: a string object that can have zero or more values, each of which must be chosen from the list of values 'value1', 'value2', ... A SET column can have a maximum of 64 members", GDA_VALUE_TYPE_STRING, NULL  },
+		{ "smallint", "", "Small integer, range is -32768 to 32767", GDA_VALUE_TYPE_SMALLINT, NULL  },
+		{ "smallint unsigned", "", "Small unsigned integer, range is 0 to 65535", GDA_VALUE_TYPE_SMALLINT, NULL  },
+		{ "text", "", "Text (up to 65535 characters)", GDA_VALUE_TYPE_STRING, NULL  },
+		{ "tinyint", "", "Tiny integer, range is -128 to 127", GDA_VALUE_TYPE_TINYINT, NULL  },
+		{ "tinyint unsigned", "", "Tiny unsigned integer, range is 0 to 255", GDA_VALUE_TYPE_TINYUINT, NULL  },
+		{ "tinyblob", "", "Tiny blob (up to 255 bytes)", GDA_VALUE_TYPE_BINARY, NULL  },
+		{ "tinytext", "", "Tiny text (up to 255 characters)", GDA_VALUE_TYPE_STRING, NULL  },		
+		{ "time", "", "Time", GDA_VALUE_TYPE_TIME, NULL  },
+		{ "timestamp", "", "Time stamp", GDA_VALUE_TYPE_TIMESTAMP, NULL  },
+		{ "varchar", "", "Variable Length Char", GDA_VALUE_TYPE_STRING, "varbinary"  },
+		{ "year", "", "Year", GDA_VALUE_TYPE_INTEGER, NULL  }
 	};
 
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
 
 	/* create the recordset */
-	recset = (GdaDataModelArray *) gda_data_model_array_new (4);
+	recset = (GdaDataModelArray *) gda_data_model_array_new (5);
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 0, _("Type"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 1, _("Owner"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 2, _("Comments"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 3, _("GDA type"));
-
+	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 4, _("Synonyms"));
+	
 	/* fill the recordset */
 	for (i = 0; i < sizeof (types) / sizeof (types[0]); i++) {
 		GList *value_list = NULL;
@@ -1021,8 +1488,9 @@ get_mysql_types (GdaConnection *cnc, GdaParameterList *params)
 		value_list = g_list_append (value_list, gda_value_new_string (types[i].owner));
 		value_list = g_list_append (value_list, gda_value_new_string (types[i].comments));
 		value_list = g_list_append (value_list, gda_value_new_type (types[i].type));
+		value_list = g_list_append (value_list, gda_value_new_string (types[i].synonyms));
 
-		gda_data_model_append_row (GDA_DATA_MODEL (recset), value_list);
+		gda_data_model_append_values (GDA_DATA_MODEL (recset), value_list);
 
 		g_list_foreach (value_list, (GFunc) gda_value_free, NULL);
 		g_list_free (value_list);
@@ -1193,7 +1661,7 @@ get_table_fields (GdaConnection *cnc, GdaParameterList *params)
 			return NULL;
 		}
 
-		gda_data_model_append_row (GDA_DATA_MODEL (recset),
+		gda_data_model_append_values (GDA_DATA_MODEL (recset),
 					   (const GList *) value_list);
 
 		g_list_foreach (value_list, (GFunc) gda_value_free, NULL);
@@ -1224,33 +1692,73 @@ gda_mysql_provider_get_schema (GdaServerProvider *provider,
 		return get_table_fields (cnc, params);
 	case GDA_CONNECTION_SCHEMA_TABLES :
 		return get_mysql_tables (cnc, params);
+	case GDA_CONNECTION_SCHEMA_VIEWS:
+		return get_mysql_views (cnc, params);
 	case GDA_CONNECTION_SCHEMA_TYPES :
 		return get_mysql_types (cnc, params);
+	case GDA_CONNECTION_SCHEMA_PROCEDURES:
+		return get_mysql_procedures (cnc, params);
 	default : break;
 	}
 
 	return NULL;
 }
 
-
-gboolean
-gda_mysql_provider_escape_string (GdaServerProvider *provider,
-				  GdaConnection *cnc,
-				  const gchar *from,
-				  gchar *to)
+gchar *
+gda_mysql_provider_value_to_sql_string (
+				GdaServerProvider *provider, /* we dont actually use this!*/
+				GdaConnection *cnc,
+				GdaValue *from)
 {
-	MYSQL *mysql;
-	GdaMysqlProvider *myprv = (GdaMysqlProvider *) provider;
+	gchar *val_str;
+	gchar *ret;
 
-	g_return_val_if_fail (GDA_IS_MYSQL_PROVIDER (myprv), FALSE);
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
 	g_return_val_if_fail (from != NULL, FALSE);
-	g_return_val_if_fail (to != NULL, FALSE);
+	
+	val_str = gda_value_stringify (from);
+	if (!val_str)
+		return NULL;
 
-	mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
-	if (!mysql) {
-		gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
-		return 0;
+	switch (from->type) {
+		case GDA_VALUE_TYPE_BIGINT :
+		case GDA_VALUE_TYPE_DOUBLE :
+		case GDA_VALUE_TYPE_INTEGER :
+		case GDA_VALUE_TYPE_NUMERIC :
+		case GDA_VALUE_TYPE_SINGLE :
+		case GDA_VALUE_TYPE_SMALLINT :
+		case GDA_VALUE_TYPE_TINYINT :
+			ret = g_strdup (val_str);
+			break;
+		case GDA_VALUE_TYPE_TIMESTAMP:
+		case GDA_VALUE_TYPE_DATE :
+		case GDA_VALUE_TYPE_TIME :
+			ret = g_strdup_printf ("\"%s\"", val_str);
+			break;
+		default:
+		{
+			MYSQL *mysql;
+			char *quoted;
+			mysql = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_MYSQL_HANDLE);
+			
+			if (!mysql) {
+				gda_connection_add_error_string (cnc, _("Invalid MYSQL handle"));
+				return NULL;
+			}
+			quoted = ret = g_malloc(strlen(val_str) * 2 + 3);
+			*quoted++ 	= '\'';
+			quoted += mysql_real_escape_string (
+					mysql, quoted, val_str, strlen (val_str));
+					
+			*quoted++     	= '\'';
+			*quoted++     	= '\0';
+			ret = g_realloc(ret, quoted - ret + 1);
+		}
+		
 	}
-	return mysql_real_escape_string (mysql, to, from, strlen (from));
+
+	g_free (val_str);
+
+	return ret;
 }
+
