@@ -1,10 +1,11 @@
 /* GNOME DB Postgres Provider
- * Copyright (C) 1998-2002 The GNOME Foundation
+ * Copyright (C) 1998-2005 The GNOME Foundation
  *
  * AUTHORS:
  *         Vivien Malerba <malerba@gnome-db.org>
  *         Rodrigo Moya <rodrigo@gnome-db.org>
  *         Gonzalo Paniagua Javier <gonzalo@gnome-db.org>
+ *         Bas Driessen <bas.driessen@xobas.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -24,6 +25,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <libgda/gda-data-model-array.h>
+#include <libgda/gda-column-index.h>
 #include "gda-postgres.h"
 #include "gda-postgres-provider.h"
 
@@ -41,22 +43,51 @@ static gboolean gda_postgres_provider_open_connection (GdaServerProvider *provid
 
 static gboolean gda_postgres_provider_close_connection (GdaServerProvider *provider,
 							GdaConnection *cnc);
+
 static const gchar *gda_postgres_provider_get_server_version (GdaServerProvider *provider,
 							      GdaConnection *cnc);
+
 static const gchar *gda_postgres_provider_get_database (GdaServerProvider *provider,
 							GdaConnection *cnc);
+
 static gboolean gda_postgres_provider_create_database (GdaServerProvider *provider,
 						       GdaConnection *cnc,
 						       const gchar *name);
+
 static gboolean gda_postgres_provider_drop_database (GdaServerProvider *provider,
 						     GdaConnection *cnc,
 						     const gchar *name);
 
+static gboolean gda_postgres_provider_create_table (GdaServerProvider *provider,
+						    GdaConnection *cnc,
+						    const gchar *table_name,
+						    const GList *attributes_list,
+						    const GList *index_list);
+
+static gboolean gda_postgres_provider_drop_table (GdaServerProvider *provider,
+						     GdaConnection *cnc,
+						     const gchar *table_name);
+
+static gboolean gda_postgres_provider_create_index (GdaServerProvider *provider,
+						    GdaConnection *cnc,
+						    const GdaDataModelIndex *index,
+						    const gchar *table_name);
+
+static gboolean gda_postgres_provider_drop_index (GdaServerProvider *provider,
+						     GdaConnection *cnc,
+						     const gchar *index_name,
+						     gboolean primary_key,
+						     const gchar *table_name);
+
 static GList *gda_postgres_provider_execute_command (GdaServerProvider *provider,
 						     GdaConnection *cnc,
 						     GdaCommand *cmd,
 						     GdaParameterList *params);
 
+static gchar *gda_postgres_provider_get_last_insert_id (GdaServerProvider *provider,
+							GdaConnection *cnc,
+							GdaDataModel *recset);
+
 static gboolean gda_postgres_provider_begin_transaction (GdaServerProvider *provider,
 							 GdaConnection *cnc,
 							 GdaTransaction *xaction);
@@ -82,6 +113,15 @@ static GdaDataModel *gda_postgres_provider_get_schema (GdaServerProvider *provid
 						       GdaConnectionSchema schema,
 						       GdaParameterList *params);
 
+static gboolean gda_postgres_provider_create_blob (GdaServerProvider *provider,
+						   GdaConnection *cnc,
+						   GdaBlob *blob);
+
+gchar *
+gda_postgres_provider_value_to_sql_string (GdaServerProvider *provider, /* we dont actually use this!*/
+					GdaConnection *cnc,
+					GdaValue *from);
+				 		 
 typedef struct {
 	gchar *col_name;
 	GdaValueType data_type;
@@ -126,12 +166,19 @@ gda_postgres_provider_class_init (GdaPostgresProviderClass *klass)
 	provider_class->get_database = gda_postgres_provider_get_database;
 	provider_class->create_database = gda_postgres_provider_create_database;
 	provider_class->drop_database = gda_postgres_provider_drop_database;
+	provider_class->create_table = gda_postgres_provider_create_table;
+	provider_class->drop_table = gda_postgres_provider_drop_table;
+	provider_class->create_index = gda_postgres_provider_create_index;
+	provider_class->drop_index = gda_postgres_provider_drop_index;
 	provider_class->execute_command = gda_postgres_provider_execute_command;
+	provider_class->get_last_insert_id = gda_postgres_provider_get_last_insert_id;
 	provider_class->begin_transaction = gda_postgres_provider_begin_transaction;
 	provider_class->commit_transaction = gda_postgres_provider_commit_transaction;
 	provider_class->rollback_transaction = gda_postgres_provider_rollback_transaction;
 	provider_class->supports = gda_postgres_provider_supports;
 	provider_class->get_schema = gda_postgres_provider_get_schema;
+	provider_class->create_blob = gda_postgres_provider_create_blob;
+	provider_class->value_to_sql_string = gda_postgres_provider_value_to_sql_string;
 }
 
 static void
@@ -208,18 +255,81 @@ postgres_name_to_gda_type (const gchar *name)
 		return GDA_VALUE_TYPE_TIME;
 	else if (!strcmp (name, "point"))
 		return GDA_VALUE_TYPE_GEOMETRIC_POINT;
+	else if (!strcmp (name, "oid"))
+		return GDA_VALUE_TYPE_BLOB;
 	else if (!strcmp (name, "bytea"))
 		return GDA_VALUE_TYPE_BINARY;
 
 	return GDA_VALUE_TYPE_STRING;
 }
 
+static gchar *
+postgres_name_from_gda_type (const GdaValueType type)
+{
+	switch (type) {
+	case GDA_VALUE_TYPE_NULL :
+		return g_strdup_printf ("text");
+	case GDA_VALUE_TYPE_BIGINT :
+		return g_strdup_printf ("int8");
+	case GDA_VALUE_TYPE_BIGUINT :
+		return g_strdup_printf ("int8");
+	case GDA_VALUE_TYPE_BINARY :
+		return g_strdup_printf ("bytea");
+	case GDA_VALUE_TYPE_BLOB :
+		return g_strdup_printf ("oid");
+	case GDA_VALUE_TYPE_BOOLEAN :
+		return g_strdup_printf ("bool");
+	case GDA_VALUE_TYPE_DATE :
+		return g_strdup_printf ("date");
+	case GDA_VALUE_TYPE_DOUBLE :
+		return g_strdup_printf ("float8");
+	case GDA_VALUE_TYPE_GEOMETRIC_POINT :
+		return g_strdup_printf ("point");
+	case GDA_VALUE_TYPE_GOBJECT :
+		return g_strdup_printf ("text");
+	case GDA_VALUE_TYPE_INTEGER :
+		return g_strdup_printf ("int4");
+	case GDA_VALUE_TYPE_LIST :
+		return g_strdup_printf ("text");
+	case GDA_VALUE_TYPE_MONEY :
+		return g_strdup_printf ("money");
+	case GDA_VALUE_TYPE_NUMERIC :
+		return g_strdup_printf ("numeric");
+	case GDA_VALUE_TYPE_SINGLE :
+		return g_strdup_printf ("float4");
+	case GDA_VALUE_TYPE_SMALLINT :
+		return g_strdup_printf ("int2");
+	case GDA_VALUE_TYPE_SMALLUINT :
+		return g_strdup_printf ("int2");
+	case GDA_VALUE_TYPE_STRING :
+		return g_strdup_printf ("varchar");
+	case GDA_VALUE_TYPE_TIME :
+		return g_strdup_printf ("time");
+	case GDA_VALUE_TYPE_TIMESTAMP :
+		return g_strdup_printf ("timestamp");
+	case GDA_VALUE_TYPE_TINYINT :
+		return g_strdup_printf ("smallint");
+	case GDA_VALUE_TYPE_TINYUINT :
+		return g_strdup_printf ("smallint");
+	case GDA_VALUE_TYPE_TYPE :
+		return g_strdup_printf ("int2");
+        case GDA_VALUE_TYPE_UINTEGER :
+		return g_strdup_printf ("int4");
+	case GDA_VALUE_TYPE_UNKNOWN :
+		return g_strdup_printf ("text");
+	default :
+		return g_strdup_printf ("text");
+	}
+
+	return g_strdup_printf ("text");
+}
+
 static int
 get_connection_type_list (GdaPostgresConnectionData *priv_td)
 {
 	GHashTable *h_table;
 	GdaPostgresTypeOid *td;
-	PGresult *pg_res, *pg_res_avoid;
+	PGresult *pg_res, *pg_res_avoid, *pg_res_anyoid = NULL;
 	gint nrows, i;
 	gchar *avoid_types = NULL;
 	GString *string;
@@ -243,7 +353,7 @@ get_connection_type_list (GdaPostgresConnectionData *priv_td)
 	}
 	else {
 		gchar *query;
-		avoid_types = "'any', 'anyarray', 'cid', 'cstring', 'int2vector', 'internal', 'language_handler', 'oid', 'oidvector', 'opaque', 'record', 'refcursor', 'regclass', 'regoper', 'regoperator', 'regproc', 'regprocedure', 'regtype', 'SET', 'smgr', 'tid', 'trigger', 'unknown', 'void', 'xid'";
+		avoid_types = "'any', 'anyarray', 'anyelement', 'cid', 'cstring', 'int2vector', 'internal', 'language_handler', 'oid', 'oidvector', 'opaque', 'record', 'refcursor', 'regclass', 'regoper', 'regoperator', 'regproc', 'regprocedure', 'regtype', 'SET', 'smgr', 'tid', 'trigger', 'unknown', 'void', 'xid'";
 
 		/* main query to fetch infos about the data types */
 		query = g_strdup_printf (
@@ -262,18 +372,26 @@ get_connection_type_list (GdaPostgresConnectionData *priv_td)
 		g_free (query);
 
 		/* query to fetch non returned data types */
-		query = g_strdup_printf ("SELECT t.oid FROM pg_catalog.pg_type t WHERE t.typname in (%s)", 
+		query = g_strdup_printf ("SELECT t.oid FROM pg_catalog.pg_type t WHERE t.typname in (%s)",
 					 avoid_types);
 		pg_res_avoid = PQexec (priv_td->pconn, query);
 		g_free (query);
+
+		/* query to fetch the oid of the 'any' data type */
+		pg_res_anyoid = PQexec (priv_td->pconn, 
+					"SELECT t.oid FROM pg_catalog.pg_type t WHERE t.typname = 'any'");
 	}
 
 	if (!pg_res || (PQresultStatus (pg_res) != PGRES_TUPLES_OK) ||
-	    !pg_res_avoid || (PQresultStatus (pg_res_avoid) != PGRES_TUPLES_OK)) {
+	    !pg_res_avoid || (PQresultStatus (pg_res_avoid) != PGRES_TUPLES_OK) ||
+	    ((priv_td->version_float >= 7.3) && 
+	     (!pg_res_anyoid || (PQresultStatus (pg_res_anyoid) != PGRES_TUPLES_OK)))) {
 		if (pg_res)
 			PQclear (pg_res);
 		if (pg_res_avoid)
 			PQclear (pg_res_avoid);
+		if (pg_res_anyoid)
+			PQclear (pg_res_anyoid);
 		return -1;
 	}
 
@@ -308,7 +426,14 @@ get_connection_type_list (GdaPostgresConnectionData *priv_td)
 	priv_td->avoid_types = avoid_types;
 	priv_td->avoid_types_oids = string->str;
 	g_string_free (string, FALSE);
-	
+
+	/* make a string of the oid of type 'any' */
+	priv_td->any_type_oid = "";
+	if (pg_res_anyoid) {
+		if (PQntuples (pg_res_anyoid) == 1) 
+			priv_td->any_type_oid = g_strdup (PQgetvalue (pg_res_anyoid, 0, 0));
+		PQclear (pg_res_anyoid);
+	}
 	return 0;
 }
 
@@ -319,7 +444,7 @@ gda_postgres_provider_get_version (GdaServerProvider *provider)
 	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
 
 	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), NULL);
-	return VERSION;
+	return PACKAGE_VERSION;
 }
 
 /* get the float version og a Postgres version which looks like:
@@ -526,8 +651,11 @@ gda_postgres_provider_close_connection (GdaServerProvider *provider, GdaConnecti
 
 	PQfinish (priv_data->pconn);
 
-	for (i = 0; i < priv_data->ntypes; i++)
+	for (i = 0; i < priv_data->ntypes; i++) {
 		g_free (priv_data->type_data[i].name);
+		g_free (priv_data->type_data[i].comments);
+		g_free (priv_data->type_data[i].owner);
+	}
 	
 	g_hash_table_destroy (priv_data->h_table);
 	g_free (priv_data->type_data);
@@ -688,6 +816,260 @@ gda_postgres_provider_drop_database (GdaServerProvider *provider,
 	return retval;
 }
 
+/* create_table handler for the GdaPostgresProvider class */
+static gboolean
+gda_postgres_provider_create_table (GdaServerProvider *provider,
+				    GdaConnection *cnc,
+				    const gchar *table_name,
+				    const GList *attributes_list,
+				    const GList *index_list)
+{
+	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+	GdaColumn *dmca;
+	GdaDataModelIndex *dmi;
+	GString *sql;
+	gint i;
+	gchar *postgres_data_type, *default_value, *references;
+	GdaValueType value_type;
+	gboolean retval;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+	g_return_val_if_fail (attributes_list != NULL, FALSE);
+
+	/* prepare the SQL command */
+	sql = g_string_new ("CREATE TABLE ");
+	g_string_append_printf (sql, "%s (", table_name);
+
+	/* step through list */
+	for (i=0; i<g_list_length ((GList *) attributes_list); i++) {
+
+		if (i>0)
+			g_string_append_printf (sql, ", ");
+			
+		dmca = (GdaColumn *) g_list_nth_data ((GList *) attributes_list, i);
+
+		/* name */	
+		g_string_append_printf (sql, "\"%s\" ", gda_column_get_name (dmca));
+
+		/* data type; force serial to be used when auto_increment is set */
+		if (gda_column_get_auto_increment (dmca) == TRUE) {
+			g_string_append_printf (sql, "serial");
+			value_type = GDA_VALUE_TYPE_INTEGER;
+		} else {
+			value_type = gda_column_get_gdatype (dmca);
+			postgres_data_type = postgres_name_from_gda_type (value_type);
+			g_string_append_printf (sql, "%s", postgres_data_type);
+			g_free(postgres_data_type);
+		}
+
+		/* size */
+		if (value_type == GDA_VALUE_TYPE_STRING)
+			g_string_append_printf (sql, "(%ld)", gda_column_get_defined_size (dmca));
+		else if (value_type == GDA_VALUE_TYPE_NUMERIC)
+			g_string_append_printf (sql, "(%ld,%ld)", 
+				gda_column_get_defined_size (dmca),
+				gda_column_get_scale (dmca));
+		
+		/* NULL */
+		if (gda_column_get_allow_null (dmca) == TRUE)
+			g_string_append_printf (sql, " NULL");
+		else
+			g_string_append_printf (sql, " NOT NULL");
+	
+		/* (primary) key */
+		if (gda_column_get_primary_key (dmca) == TRUE)
+			g_string_append_printf (sql, " PRIMARY KEY");
+		else
+			if (gda_column_get_unique_key (dmca) == TRUE)
+				g_string_append_printf (sql, " UNIQUE");
+
+		/* default value (in case of string, user needs to add "'" around the field) */
+		if (gda_column_get_default_value (dmca) != NULL) {
+			default_value = gda_value_stringify ((GdaValue *) gda_column_get_default_value (dmca));
+			if ((default_value != NULL) && (*default_value != '\0'))
+				g_string_append_printf (sql, " DEFAULT %s", default_value);
+		}
+
+		/* any additional parameters */			
+		if (gda_column_get_references (dmca) != NULL) {
+			references = (gchar *) gda_column_get_references (dmca);
+			if ((references != NULL) && (*references != '\0'))
+				g_string_append_printf (sql, " %s", references);
+		}
+	}
+
+	/* finish the SQL command */
+	g_string_append_printf (sql, ")");
+
+	/* execute sql command */
+	retval = gda_postgres_provider_single_command (pg_prv, cnc, sql->str);
+
+	/* clean up */
+	g_string_free (sql, TRUE);
+
+	/* Create indexes */
+	if (index_list != NULL) {
+
+		/* step through index_list */
+		for (i=0; i<g_list_length ((GList *) index_list); i++) {
+
+			/* get index */
+			dmi = (GdaDataModelIndex *) g_list_nth_data ((GList *) index_list, i);
+
+			/* create index */
+			retval = gda_postgres_provider_create_index (provider, cnc, dmi, table_name);
+			if (retval == FALSE)
+				return FALSE;
+		}
+	}
+
+	/* return */
+	return retval;
+}
+
+/* drop_table handler for the GdaPostgresProvider class */
+static gboolean
+gda_postgres_provider_drop_table (GdaServerProvider *provider,
+				  GdaConnection *cnc,
+				  const gchar *table_name)
+{
+	gboolean retval;
+	gchar *sql;
+
+	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	sql = g_strdup_printf ("DROP TABLE %s", table_name);
+
+	retval = gda_postgres_provider_single_command (pg_prv, cnc, sql);
+	g_free (sql);
+
+	return retval;
+}
+
+/* create_index handler for the GdaPostgresProvider class */
+static gboolean
+gda_postgres_provider_create_index (GdaServerProvider *provider,
+				    GdaConnection *cnc,
+				    const GdaDataModelIndex *index,
+				    const gchar *table_name)
+{
+	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+	GdaColumnIndex *dmcia;
+	GString *sql;
+	GList *col_list;
+	gchar *index_name, *col_ref, *idx_ref;
+	gint i;
+	gboolean retval;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (index != NULL, FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	/* the 'CREATE INDEX' command in PostgreSQL is limited to
+	secondary non-primary indices only. Therefore use 'ALTER TABLE'
+	for PRIMARY and UNIQUE */
+
+	/* prepare the SQL command */
+	sql = g_string_new (" ");
+
+	/* get index name */
+	index_name = (gchar *) gda_data_model_index_get_name ((GdaDataModelIndex *) index);
+
+	/* determine type of index (PRIMARY, UNIQUE or INDEX) */
+	if (gda_data_model_index_get_primary_key ((GdaDataModelIndex *) index) == TRUE)
+		g_string_append_printf (sql, "ALTER TABLE %s ADD CONSTRAINT %s PRIMARY KEY (", table_name, index_name);
+	else if (gda_data_model_index_get_unique_key ((GdaDataModelIndex *) index) == TRUE)
+		g_string_append_printf (sql, "ALTER TABLE %s ADD CONSTRAINT %s UNIQUE (", table_name, index_name);
+	else
+		g_string_append_printf (sql, "CREATE INDEX %s ON %s (", index_name, table_name);
+
+	/* get the list of index columns */
+	col_list = gda_data_model_index_get_column_index_list ((GdaDataModelIndex *) index);
+
+	/* step through list */
+	for (i = 0; i < g_list_length ((GList *) col_list); i++) {
+
+		if (i > 0)
+			g_string_append_printf (sql, ", ");
+			
+		dmcia = (GdaColumnIndex *) g_list_nth_data ((GList *) col_list, i);
+
+		/* name */	
+		g_string_append_printf (sql, "\"%s\" ", gda_column_index_get_column_name (dmcia));
+
+		/* any additional column parameters */
+		if (gda_column_index_get_references (dmcia) != NULL) {
+			col_ref = (gchar *) gda_column_index_get_references (dmcia);
+			if ((col_ref != NULL) && (*col_ref != '\0'))
+				g_string_append_printf (sql, " %s ", col_ref);
+		}
+	}
+
+	/* finish the SQL column(s) section command */
+	g_string_append_printf (sql, ")");
+
+	/* any additional index parameters */
+	if (gda_data_model_index_get_references ((GdaDataModelIndex *) index) != NULL) {
+		idx_ref = (gchar *) gda_data_model_index_get_references ((GdaDataModelIndex *) index);
+		if ((idx_ref != NULL) && (*idx_ref != '\0'))
+			g_string_append_printf (sql, " %s ", idx_ref);
+	}
+
+	/* execute sql command */
+	retval = gda_postgres_provider_single_command (pg_prv, cnc, sql->str);
+
+	/* clean up */
+	g_string_free (sql, TRUE);
+
+	return retval;
+}
+
+/* drop_index handler for the GdaPostgresProvider class */
+static gboolean
+gda_postgres_provider_drop_index (GdaServerProvider *provider,
+				  GdaConnection *cnc,
+				  const gchar *index_name,
+				  gboolean primary_key,
+				  const gchar *table_name)
+{
+	gboolean retval = FALSE, retval_temp;
+	gchar *sql_alter;
+	gchar *sql_drop;
+
+	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (index_name != NULL, FALSE);
+	g_return_val_if_fail (table_name != NULL, FALSE);
+
+	/* In PostgreSQL PRIMARY can only and UNIQUE can as well be defined as
+	   constraints. Therefore use 2 methods for removal */
+	sql_alter = g_strdup_printf ("ALTER TABLE %s DROP CONSTRAINT %s", table_name, index_name);
+	sql_drop  = g_strdup_printf ("DROP INDEX %s", index_name);
+
+	retval_temp = gda_postgres_provider_single_command (pg_prv, cnc, sql_alter);
+	if (retval_temp == TRUE)
+		retval = TRUE;
+
+	retval_temp = gda_postgres_provider_single_command (pg_prv, cnc, sql_drop);
+	if (retval_temp == TRUE)
+		retval = TRUE;
+
+	/* clean up */
+	g_free (sql_alter);
+	g_free (sql_drop);
+
+	return retval;
+}
+
 /* execute_command handler for the GdaPostgresProvider class */
 static GList *
 gda_postgres_provider_execute_command (GdaServerProvider *provider,
@@ -722,17 +1104,89 @@ gda_postgres_provider_execute_command (GdaServerProvider *provider,
 	return reclist;
 }
 
+static gchar *
+gda_postgres_provider_get_last_insert_id (GdaServerProvider *provider,
+					  GdaConnection *cnc,
+					  GdaDataModel *recset)
+{
+	Oid oid;
+	PGresult *pgres;
+	GdaPostgresConnectionData *priv_data;
+	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), NULL);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
+	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
+
+	priv_data = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_POSTGRES_HANDLE);
+	if (!priv_data) {
+		gda_connection_add_error_string (cnc, _("Invalid PostgreSQL handle"));
+		return NULL;
+	}
+
+	/* get the PQresult from the recordset */
+	pgres = gda_postgres_recordset_get_pgresult (GDA_POSTGRES_RECORDSET (recset));
+	if (pgres) {
+		oid = PQoidValue (pgres);
+		if (oid != InvalidOid)
+			return g_strdup_printf ("%d", oid);
+	}
+
+	return NULL;
+}
+
 static gboolean 
 gda_postgres_provider_begin_transaction (GdaServerProvider *provider,
 				         GdaConnection *cnc,
 					 GdaTransaction *xaction)
 {
+        gboolean result;
 	GdaPostgresProvider *pg_prv = (GdaPostgresProvider *) provider;
+	GdaPostgresConnectionData *priv_data;
+
+	gchar * write_option=NULL;
+	gchar * isolation_level=NULL;
 
 	g_return_val_if_fail (GDA_IS_POSTGRES_PROVIDER (pg_prv), FALSE);
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+	g_return_val_if_fail (GDA_IS_TRANSACTION (xaction), FALSE);
 
-	return gda_postgres_provider_single_command (pg_prv, cnc, "BEGIN"); 
+	priv_data = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_POSTGRES_HANDLE);
+	if (priv_data->version_float >= 6.5){
+	        if (gda_connection_get_options (cnc) & GDA_CONNECTION_OPTIONS_READ_ONLY) {
+	                if (priv_data->version_float >= 7.4){
+		                write_option = "READ ONLY";
+           	        } 
+                        else {
+		                gda_connection_add_error_string (cnc, _("Transactions are not supported in read-only mode"));
+				return FALSE;
+			}
+		}
+		switch (gda_transaction_get_isolation_level (xaction)) {
+		case GDA_TRANSACTION_ISOLATION_READ_COMMITTED :
+		        isolation_level = g_strconcat ("SET TRANSACTION ISOLATION LEVEL READ COMMITTED ", write_option, NULL);
+		        break;
+		case GDA_TRANSACTION_ISOLATION_READ_UNCOMMITTED :
+		        gda_connection_add_error_string (cnc, _("Transactions are not supported in read uncommitted isolation level"));
+		        return FALSE;
+		case GDA_TRANSACTION_ISOLATION_REPEATABLE_READ :
+		        gda_connection_add_error_string (cnc, _("Transactions are not supported in repeatable read isolation level"));
+		        return FALSE;
+		case GDA_TRANSACTION_ISOLATION_SERIALIZABLE :
+		        isolation_level = g_strconcat ("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE ", write_option, NULL);
+		        break;
+		default: 
+		        isolation_level = NULL;
+		}
+	}
+
+	result = gda_postgres_provider_single_command (pg_prv, cnc, "BEGIN"); 
+	if (result&&isolation_level != NULL) {
+	        result=gda_postgres_provider_single_command (pg_prv, cnc, isolation_level) ;
+	} 
+	g_free(isolation_level);
+
+	return result;
 }
 
 static gboolean 
@@ -812,6 +1266,7 @@ static gboolean gda_postgres_provider_supports (GdaServerProvider *provider,
 	case GDA_CONNECTION_FEATURE_TRIGGERS :
 	case GDA_CONNECTION_FEATURE_USERS :
 	case GDA_CONNECTION_FEATURE_VIEWS :
+	case GDA_CONNECTION_FEATURE_BLOBS :
 		return TRUE;
 	case GDA_CONNECTION_FEATURE_NAMESPACES :
 		priv_data = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_POSTGRES_HANDLE);
@@ -894,12 +1349,12 @@ gda_postgres_fill_procs_data (GdaDataModelArray *recset,
 			 "AND p.prorettype <> 'pg_catalog.cstring'::pg_catalog.regtype "
 			 "AND p.proargtypes[0] <> 'pg_catalog.cstring'::pg_catalog.regtype "
 			 "AND t.oid in (SELECT ty.oid FROM pg_catalog.pg_type ty WHERE ty.typrelid = 0 AND "
-			 "ty.typname !~ '^_' AND ty.oid not in (%s)) "
+			 "ty.typname !~ '^_' AND (ty.oid not in (%s) OR ty.oid = '%s')) "
 			 "AND p.proretset = 'f' "
 			 "AND NOT p.proisagg "
 			 "AND pg_catalog.pg_function_is_visible(p.oid) "
 			 "ORDER BY proname, pronargs",
-			 priv_data->avoid_types_oids);
+			 priv_data->avoid_types_oids, priv_data->any_type_oid);
 
 	pg_res = PQexec(priv_data->pconn, query);
 	g_free (query);
@@ -955,7 +1410,7 @@ gda_postgres_fill_procs_data (GdaDataModelArray *recset,
 			while (ptr && *ptr && insert) {
 				const GdaPostgresTypeOid *typeoid = NULL;
 
-				if (*ptr == '0') {
+				if (!strcmp (ptr, priv_data->any_type_oid)) {
 					thevalue = "-";
 				}
 				else {
@@ -972,7 +1427,7 @@ gda_postgres_fill_procs_data (GdaDataModelArray *recset,
 					if (!gstr) 
 						gstr = g_string_new (thevalue);
 					else
-						g_string_append_printf (gstr, " %s", thevalue);
+						g_string_append_printf (gstr, ",%s", thevalue);
 				}
 					
 				ptr = strtok (NULL, " ");
@@ -1116,28 +1571,50 @@ get_postgres_types (GdaConnection *cnc, GdaParameterList *params)
 	GdaDataModelArray *recset;
 	GdaPostgresConnectionData *priv_data;
 	gint i;
+	static GHashTable *synonyms = NULL;
+
+	if (!synonyms) {
+		synonyms = g_hash_table_new (g_str_hash, g_str_equal);
+
+		g_hash_table_insert (synonyms, "int4", "int,integer");
+		g_hash_table_insert (synonyms, "int8", "bigint");
+		g_hash_table_insert (synonyms, "serial8", "bigserial");
+		g_hash_table_insert (synonyms, "varbit", "bit varying");
+		g_hash_table_insert (synonyms, "bool", "boolean");
+		g_hash_table_insert (synonyms, "varchar", "character varying");
+		g_hash_table_insert (synonyms, "char", "character");
+		g_hash_table_insert (synonyms, "float8", "double precision");
+		g_hash_table_insert (synonyms, "numeric", "decimal");
+		g_hash_table_insert (synonyms, "float4", "real");
+		g_hash_table_insert (synonyms, "int2", "smallint");
+		g_hash_table_insert (synonyms, "serial4", "serial");
+	}
 
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
 
 	/* create the recordset */
-	recset = GDA_DATA_MODEL_ARRAY (gda_data_model_array_new (4));
+	recset = GDA_DATA_MODEL_ARRAY (gda_data_model_array_new (5));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 0, _("Type"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 1, _("Owner"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 2, _("Comments"));
 	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 3, _("GDA type"));
+	gda_data_model_set_column_title (GDA_DATA_MODEL (recset), 4, _("Synonyms"));
 
 	/* fill the recordset */
 	priv_data = g_object_get_data (G_OBJECT (cnc), OBJECT_DATA_POSTGRES_HANDLE);
 
 	for (i = 0; i < priv_data->ntypes; i++) {
 		GList *value_list = NULL;
+		gchar *syn;
 
 		value_list = g_list_append (value_list, gda_value_new_string (priv_data->type_data[i].name));
 		value_list = g_list_append (value_list, gda_value_new_string (priv_data->type_data[i].owner));
 		value_list = g_list_append (value_list, gda_value_new_string (priv_data->type_data[i].comments));
 		value_list = g_list_append (value_list, gda_value_new_type (priv_data->type_data[i].type));
+		syn = g_hash_table_lookup (synonyms, priv_data->type_data[i].name);
+		value_list = g_list_append (value_list, gda_value_new_string (syn));
 
-		gda_data_model_append_row (GDA_DATA_MODEL (recset), value_list);
+		gda_data_model_append_values (GDA_DATA_MODEL (recset), value_list);
 
 		g_list_foreach (value_list, (GFunc) gda_value_free, NULL);
 		g_list_free (value_list);
@@ -1268,24 +1745,25 @@ get_postgres_aggregates (GdaConnection *cnc, GdaParameterList *params)
 
 			query = g_strdup_printf (
                                   "SELECT p.proname, p.oid, u.usename, pg_catalog.obj_description(p.oid),"
+				  "typo.typname,"
 				  "CASE p.proargtypes[0] "
 				  "WHEN 'pg_catalog.\"any\"'::pg_catalog.regtype "
 				  "THEN CAST('-' AS pg_catalog.text) "
 				  "ELSE typi.typname "
 				  "END,"
-				  "typo.typname, NULL "
+				  "NULL "
 				  "FROM pg_catalog.pg_proc p, pg_catalog.pg_user u, pg_catalog.pg_namespace n, "
 				  "pg_catalog.pg_type typi, pg_catalog.pg_type typo "
 				  "WHERE u.usesysid=p.proowner "
 				  "AND n.oid = p.pronamespace "
 				  "AND p.prorettype = typo.oid "
-				  "AND typo.oid NOT IN (%s) "
+				  "AND (typo.oid NOT IN (%s) OR typo.oid='%s') "
 				  "AND p.proargtypes[0] = typi.oid "
-				  "AND typi.oid NOT IN (%s) "
+				  "AND (typi.oid NOT IN (%s) OR typi.oid='%s') "
 				  "AND p.proisagg "
 				  "AND pg_catalog.pg_function_is_visible(p.oid) "
-				  "ORDER BY 2", priv_data->avoid_types_oids, 
-				  priv_data->avoid_types_oids);
+				  "ORDER BY 2", priv_data->avoid_types_oids, priv_data->any_type_oid,
+				  priv_data->avoid_types_oids, priv_data->any_type_oid);
 			reclist = process_sql_commands (NULL, cnc, query, GDA_COMMAND_OPTION_STOP_ON_ERRORS);
 			g_free (query);
 		}
@@ -1458,6 +1936,11 @@ gda_postgres_index_type (gint colnum, GList *idx_list, IdxType type)
 	return FALSE;
 }
 
+#ifdef LIBGDA_WIN32
+/* this is a ming thing.. */
+#define strtok_r(a,b,c) strtok(a,b)
+
+#endif
 
 /* Converts a single dimension array in the form of '{AA,BB,CC}' to a list of
  * strings (here ->"AA"->"BB"->"CC) which must be freed by the caller
@@ -1645,19 +2128,19 @@ gda_postgres_fill_md_data (const gchar *tblname, GdaDataModelArray *recset,
 
 	if (priv_data->version_float < 7.3)
 		query = g_strdup_printf (
-			"(SELECT a.attname, b.typname, a.atttypmod, b.typlen, not a.attnotnull, d.adsrc, "
+			"(SELECT a.attname, b.typname, a.atttypmod, b.typlen, a.attnotnull, d.adsrc, "
 			"a.attnum FROM pg_class c, pg_attribute a, pg_type b, pg_attrdef d "
 			"WHERE c.relname = '%s' AND a.attnum > 0 AND "
 			"a.attrelid = c.oid and b.oid = a.atttypid AND "
 			"a.atthasdef = 't' and d.adrelid=c.oid and d.adnum=a.attnum) "
-			"UNION (SELECT a.attname, b.typname, a.atttypmod, b.typlen, not a.attnotnull, NULL, "
+			"UNION (SELECT a.attname, b.typname, a.atttypmod, b.typlen, a.attnotnull, NULL, "
 			"a.attnum FROM pg_class c, pg_attribute a, pg_type b "
 			  "WHERE c.relname = '%s' AND a.attnum > 0 AND "
 			"a.attrelid = c.oid and b.oid = a.atttypid AND a.atthasdef = 'f') ORDER BY 7",
 			tblname, tblname);
 	else 
 		query = g_strdup_printf (
-			"SELECT a.attname, t.typname, a.atttypmod, t.typlen, not a.attnotnull, d.adsrc, a.attnum "
+			"SELECT a.attname, t.typname, a.atttypmod, t.typlen, a.attnotnull, pg_get_expr (d.adbin, c.oid), a.attnum "
 			"FROM pg_catalog.pg_class c "
 			"LEFT JOIN pg_catalog.pg_attribute a ON (a.attrelid = c.oid) "
 			"FULL JOIN pg_catalog.pg_attrdef d ON (a.attnum = d.adnum AND d.adrelid=c.oid) "
@@ -1792,7 +2275,7 @@ add_g_list_row (gpointer data, gpointer user_data)
 	GList *rowlist = data;
 	GdaDataModelArray *recset = user_data;
 
-	gda_data_model_append_row (GDA_DATA_MODEL (recset), rowlist);
+	gda_data_model_append_values (GDA_DATA_MODEL (recset), rowlist);
 	g_list_foreach (rowlist, (GFunc) gda_value_free, NULL);
 	g_list_free (rowlist);
 }
@@ -2034,3 +2517,67 @@ gda_postgres_provider_get_schema (GdaServerProvider *provider,
 	return NULL;
 }
 
+static gboolean gda_postgres_provider_create_blob (GdaServerProvider *provider,
+						   GdaConnection *cnc,
+						   GdaBlob *blob)
+{
+	g_return_val_if_fail (GDA_IS_SERVER_PROVIDER (provider), FALSE);
+	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
+
+	return gda_postgres_blob_create (blob, cnc);
+}
+
+gchar *
+gda_postgres_provider_value_to_sql_string (
+				GdaServerProvider *provider, /* we dont actually use this!*/
+				GdaConnection *cnc,
+				GdaValue *from)
+{
+	gchar *val_str;
+	gchar *ret;
+	
+	g_return_val_if_fail (from != NULL, NULL);
+
+	val_str = gda_value_stringify (from);
+	if (!val_str)
+		return NULL;
+
+	switch (from->type) {
+		case GDA_VALUE_TYPE_BIGINT :
+		case GDA_VALUE_TYPE_DOUBLE :
+		case GDA_VALUE_TYPE_INTEGER :
+		case GDA_VALUE_TYPE_NUMERIC :
+		case GDA_VALUE_TYPE_SINGLE :
+		case GDA_VALUE_TYPE_SMALLINT :
+		case GDA_VALUE_TYPE_TINYINT :
+			ret = g_strdup (val_str);
+			break;
+		case GDA_VALUE_TYPE_TIMESTAMP:
+		case GDA_VALUE_TYPE_DATE :
+		case GDA_VALUE_TYPE_TIME :
+			ret = g_strdup_printf ("\"%s\"", val_str);
+			break;
+		case GDA_VALUE_TYPE_BINARY:
+		{
+			int qlen;
+			char *quoted;
+			quoted = PQescapeBytea(val_str, (size_t) strlen(val_str), &qlen);
+			ret = g_strdup_printf ("\"%s\"", quoted);
+			PQfreemem(quoted);
+			break;
+		}
+		default :
+		{
+			char *quoted;
+			quoted = ret = g_malloc(strlen(val_str) * 2 + 3);
+			*quoted++     	= '\'';
+			quoted += PQescapeString (quoted, val_str, (size_t) strlen (val_str));
+			*quoted++     	= '\'';
+			*quoted++     	= '\0';
+			ret = g_realloc(ret, quoted - ret + 1);
+		}
+	}
+	g_free (val_str);
+
+	return ret;
+}
